#include <stdio.h>
#include <stdlib.h>
#include "db_compiler.h"
#include "../common/db_mdio.h"

typedef unsigned long uint32_t;

#define CHAMELEON_PIC_LOAD_ADDRESS  0x00008000
#define XGS_PIC_LOAD_ADDRESS        0x00008000
#define PIC_BAUD_RATE               115200

#define PIC_COMMAND_NACK        0x00
#define PIC_COMMAND_ACK         0x01
#define PIC_COMMAND_READ_PM     0x02
#define PIC_COMMAND_WRITE_PM    0x03
#define PIC_COMMAND_READ_EE     0x04
#define PIC_COMMAND_WRITE_EE    0x05
#define PIC_COMMAND_READ_CM     0x06
#define PIC_COMMAND_WRITE_CM    0x07
#define PIC_COMMAND_RESET       0x08
#define PIC_COMMAND_READ_ID     0x09

#define PIC_FLASH_ROW_SIZE      512

static int PicWriteImage(const char *port, const char *file, uint32_t addr);
static int PicWriteBlock(uint32_t addr, uint8_t *buf, int len);
static void PicReset(void);

#define CHAMELEON_AVR_LOAD_ADDRESS  0x6000
#define XGS_AVR_LOAD_ADDRESS        0x6000
#define AVR_BAUD_RATE               57600

#define AVR_ENTER_PROGMODE      'P' /* EOC */
#define AVR_LEAVE_PROGMODE      'Q' /* EOC */
#define AVR_SET_ADDRESS         'U' /* AL AH EOC */
#define AVR_WRITE_PAGE          'd' /* CH CL TYPE data... EOC */
#define AVR_TYPE_FLASH          'F'
#define AVR_TYPE_EEPROM         'E'
#define AVR_END_OF_COMMAND      ' '
#define AVR_INSYNC              0x14
#define AVR_OK                  0x10

#define AVR_FLASH_PAGE_SIZE     128

static int AvrWriteImage(const char *port, const char *file, uint16_t addr);
static int AvrStartLoad(void);
static int AvrEndLoad(void);
static int AvrWriteBlock(uint16_t addr, uint8_t *buf, int len);
static int AvrGetResponse(void);

int WriteImage(const char *port, int boardType, const char *file)
{
    int sts;
    switch (boardType) {
    case BOARD_CHAMELEON_AVR:
        sts = AvrWriteImage(port, file, CHAMELEON_AVR_LOAD_ADDRESS);
        break;
    case BOARD_CHAMELEON_PIC:
        sts = PicWriteImage(port, file, CHAMELEON_PIC_LOAD_ADDRESS);
        break;
    case BOARD_XGS_AVR:
        sts = AvrWriteImage(port, file, XGS_AVR_LOAD_ADDRESS);
        break;
    case BOARD_XGS_PIC:
        sts = PicWriteImage(port, file, XGS_PIC_LOAD_ADDRESS);
        break;
    default:
        sts = VMFALSE;
        break;
    }
    return sts;
}

static int PicWriteImage(const char *port, const char *file, uint32_t addr)
{
    uint8_t buf[PIC_FLASH_ROW_SIZE * 2];
    int byte, i = 0;
    FILE *fp;

    if (!md_init(port, PIC_BAUD_RATE))
        return VMFALSE;

    if (!(fp  = fopen(file, "rb")))
        goto fail;
    
    md_reset();

    while ((byte = getc(fp)) != EOF) {
        buf[i++] = byte;
        if (i >= sizeof(buf)) {
            if (PicWriteBlock(addr, buf, sizeof(buf)) != PIC_COMMAND_ACK)
                goto fail;
            addr += PIC_FLASH_ROW_SIZE * 2;
            i = 0;
        }
    }
    
    fclose(fp);
    fp = NULL;

    if (i > 0) {
        while (i < sizeof(buf))
            buf[i++] = 0;
        if (PicWriteBlock(addr, buf, sizeof(buf)) != PIC_COMMAND_ACK)
            goto fail;
    }

    PicReset();

    md_done();

    return VMTRUE;

fail:
    md_done();
    if (fp)
        fclose(fp);
    return VMFALSE;
}

static int PicWriteBlock(uint32_t addr, uint8_t *buf, int len)
{
    uint8_t cmd[4 + PIC_FLASH_ROW_SIZE * 3];
    int i, j;
    cmd[0] = PIC_COMMAND_WRITE_PM;
	cmd[1] = (uint8_t) addr;
	cmd[2] = (uint8_t)(addr >>  8);
	cmd[3] = (uint8_t)(addr >> 16);
    for (i = 0, j = 4; i < len; ++i) {
        cmd[j++] = buf[i];
        if ((i & 1) == 1)
            cmd[j++] = '\0';
    }
    md_write(cmd, sizeof(cmd));
    return md_get(1);
}

static void PicReset(void)
{
    uint8_t cmd[1];
    // The following was originally PIC_COMMAND_RESET but that command
    // writes some configuration bytes before doing the reset and that
    // was trashing the bootloader for some downloads. The PIC_COMMAND_NAK
    // command just does a reset and nothing else.
    cmd[0] = PIC_COMMAND_NACK;
    md_write(cmd, sizeof(cmd));
}

static int AvrWriteImage(const char *port, const char *file, uint16_t addr)
{
    uint8_t buf[AVR_FLASH_PAGE_SIZE];
    int byte, i = 0;
    FILE *fp;

    if (!md_init(port, AVR_BAUD_RATE))
        return VMFALSE;

    if (!(fp  = fopen(file, "rb")))
        goto fail;
    
    md_reset();

    if (!AvrStartLoad())
        goto fail;

    while ((byte = getc(fp)) != EOF) {
        buf[i++] = byte;
        if (i >= sizeof(buf)) {
            if (!AvrWriteBlock(addr, buf, sizeof(buf)))
                goto fail;
            addr += AVR_FLASH_PAGE_SIZE;
            i = 0;
        }
    }
    
    fclose(fp);
    fp = NULL;

    if (i > 0) {
        while (i < sizeof(buf))
            buf[i++] = 0;
        if (!AvrWriteBlock(addr, buf, sizeof(buf)))
            goto fail;
    }

    if (!AvrEndLoad())
        goto fail;

    md_done();

    return VMTRUE;

fail:
    md_done();
    if (fp)
        fclose(fp);
    return VMFALSE;
}

static int AvrStartLoad(void)
{
    char cmd[4];
    cmd[0] = AVR_ENTER_PROGMODE;
    cmd[1] = AVR_END_OF_COMMAND;
    md_write(cmd, 2);
    return AvrGetResponse();
}

static int AvrEndLoad(void)
{
    char cmd[2];
    cmd[0] = AVR_LEAVE_PROGMODE;
    cmd[1] = AVR_END_OF_COMMAND;
    md_write(cmd, sizeof(cmd));
    return AvrGetResponse();
}

static int AvrWriteBlock(uint16_t addr, uint8_t *buf, int len)
{
    uint16_t wordAddr = addr >> 1;
    char cmd[4];
    cmd[0] = AVR_SET_ADDRESS;
    cmd[1] = (uint8_t)wordAddr;
    cmd[2] = (uint8_t)(wordAddr >> 8);
    cmd[3] = AVR_END_OF_COMMAND;
    md_write(cmd, 4);
    if (!AvrGetResponse())
        return VMFALSE;
    cmd[0] = AVR_WRITE_PAGE;
    cmd[1] = (uint8_t)(len >> 8);
    cmd[2] = (uint8_t)len;
    cmd[3] = AVR_TYPE_FLASH;
    md_write(cmd, 4);
    md_write(buf, len);
    cmd[0] = AVR_END_OF_COMMAND;
    md_write(cmd, 1);
    return AvrGetResponse();
}

static int AvrGetResponse(void)
{
    int sts = md_get(1) == AVR_INSYNC && md_get(1) == AVR_OK;
    if (!sts)
        VM_printf(" -- failed\n");
    return sts;
}


